-----------------------------------------------------------------------------------------------
-- Client Lua Script for CombatEvents
-- Copyright (c) NCsoft. All rights reserved
-----------------------------------------------------------------------------------------------

require "Spell"
require "Window"

--[[ Events

	CombatEvents_Loaded				used by addons checking to ensure CombatEvents is loaded
											tEventArgs,
												version, number
												
	CE_UnitCreated					Unit Creation event with health, shields, absorb and casting data
											tEventArgs,
												unit, unitObject
												unitId, unitId
												name, string
												type, string
												isCasting, boolean
												health, table (same data as CE_UnitVitalsChanged)
												shield, table (same data as CE_UnitVitalsChanged)
												absorb, table (same data as CE_UnitVitalsChanged)
												cast, table (same data as CE_UnitCastStateChanged)
													
	CE_UnitDestroyed				Unit Removal event with unitId
											tEventArgs,
												unitId, userId
												
	CE_UnitVitalsChanged			Unit Vital event, returns health, shields absorb tables with current and max values if changed
											tEventArgs,
												unit, unitObject
												unitId, unitId
												health, table (current, max)
												shield, table (current, max)
												absorb, table (current, max)
											
	CE_UnitCastStateChanged			Unit Casting event, returns is_casting and spell details at cast time and cast end
											tEventArgs,
												unit, unitObject
												unitId, unitId
												is_casting, boolean
												type, enum
												name, string
												pct, float
												duration, float
												elapsed, float
--]]
 
-----------------------------------------------------------------------------------------------
-- CombatEvents Module Definition
-----------------------------------------------------------------------------------------------
local CombatEvents = {} 
 
-----------------------------------------------------------------------------------------------
-- Constants
-----------------------------------------------------------------------------------------------
local nVersion = 1

local kRequiredCvars = {
	"cmbtlog.disableOtherPlayers",
	"cmbtlog.disableAbsorption",
	"cmbtlog.disableCCState",
	"cmbtlog.disableDamage",
	"cmbtlog.disableDeflect",
	"cmbtlog.disableDelayDeath",
	"cmbtlog.disableDispel",
	"cmbtlog.disableFallingDamage",
	"cmbtlog.disableHeal",
	"cmbtlog.disableImmunity",
	"cmbtlog.disableInterrupted",
	"cmbtlog.disableModifyInterruptArmor",
	"cmbtlog.disableTransference",
	"cmbtlog.disableVitalModifier",
	"cmbtlog.disableDeath",
}


local kUpdateTimers = {
	{				id="auto",		label="Auto"},		-- Engine Controlled
	{fps=60,		id="accurate",	label="Accurate"},	-- 60FPS
	{fps=30,		id="veryfast",	label="Very Fast"},	-- 30FPS
	{fps=20,		id="fast",		label="Fast"},		-- 20FPS
	{fps=10,		id="normal",	label="Normal"},	-- 10FPS
	{fps=5,			id="slow",		label="Slow"},		-- 5FPS
	{fps=2,			id="veryslow",	label="Very Slow"},	-- 2FPS
}

local kBlacklistedUnitTypes = {
	["Collectible"] = true,
	["PinataLoot"] = true,

}
 
-----------------------------------------------------------------------------------------------
-- Initialization
-----------------------------------------------------------------------------------------------
function CombatEvents:new(o)
    o = o or {}
    setmetatable(o, self)
    self.__index = self 

    -- initialize variables here
	o.unitPreload = {}
	
	o.tUnits = {}
	o.tUnitIndex = {}
	
	o.strRate = nil

    return o
end

function CombatEvents:Init()
	local bHasConfigureFunction = true
	local strConfigureButtonText = "Combat Events"
	local tDependencies = {
		-- "UnitOrPackageName",
	}
    Apollo.RegisterAddon(self, bHasConfigureFunction, strConfigureButtonText, tDependencies)
end
 

-----------------------------------------------------------------------------------------------
-- CombatEvents OnLoad
-----------------------------------------------------------------------------------------------
function CombatEvents:OnLoad()
    -- load our form file
	self.xmlDoc = XmlDoc.CreateFromFile("CombatEvents.xml")
	self.xmlDoc:RegisterCallback("OnDocLoaded", self)
	
	-- Check Cvars and enable, required for use!
	for _,cvar in ipairs(kRequiredCvars) do
		local enabled = Apollo.GetConsoleVariable(cvar)
		if enabled then
			Apollo.SetConsoleVariable(cvar, false)
		end
	end
		
	Apollo.RegisterEventHandler("UnitCreated", 					"OnPreloadUnitCreated", 					self)
end

-----------------------------------------------------------------------------------------------
-- CombatEvents OnDocLoaded
-----------------------------------------------------------------------------------------------
function CombatEvents:OnDocLoaded()
	Apollo.RemoveEventHandler("UnitCreated", self)
	if self.xmlDoc ~= nil and self.xmlDoc:IsLoaded() then
	    self.wndOptions = Apollo.LoadForm(self.xmlDoc, "OptionsForm", nil, self)
		if self.wndOptions == nil then
			Apollo.AddAddonErrorText(self, "CombatEvents Options Window has failed to load.")
			return nil
		end
		self.wndOptions:Show(false, true)
		
		self.wndRates = {}
		for i, updateInfo in ipairs(kUpdateTimers) do
			self.wndRates[i] = self.wndOptions:FindChild("Body:UpdateRate:DropToggleFPS:ChoiceContainer:FPS"..i)
			local button = self.wndRates[i]
			button:SetData(updateInfo.id)
			button:SetText(updateInfo.label)
			
			
			
			-- Create Timer
			if updateInfo.fps then
				button:FindChild("Label_Rate"):SetText("("..updateInfo.fps.." FPS)")
				
				Apollo.RegisterTimerHandler("UpdateTimer_"..updateInfo.id, "OnUnitUpdate", self)
				Apollo.CreateTimer("UpdateTimer_"..updateInfo.id, 1/updateInfo.fps, true)
				Apollo.StopTimer("UpdateTimer_"..updateInfo.id)
			else
			button:FindChild("Label_Rate"):SetText("(Engine)")
			end
		end
		self.wndOptions:FindChild("Body:Version"):SetText("Version: "..nVersion)

		
		-- Unit Discovery
		Apollo.RegisterEventHandler("UnitCreated", 						"OnUnitCreated", 						self)
		Apollo.RegisterEventHandler("UnitDestroyed", 					"OnUnitDestroyed",						self)

		-- Vitals
		Apollo.RegisterEventHandler("CombatLogDamage", 					"OnUnitHealthChanged",					self)
		Apollo.RegisterEventHandler("CombatLogHeal", 					"OnUnitHealthChanged",					self)
		Apollo.RegisterEventHandler("CombatLogFallingDamage", 			"OnUnitHealthChanged",					self)
		Apollo.RegisterEventHandler("CombatLogLifeSteal", 				"OnUnitHealthChanged",					self)
		Apollo.RegisterEventHandler("CombatLogTransference", 			"OnUnitHealthChanged",					self)
		Apollo.RegisterEventHandler("CombatLogVitalModifier", 			"OnUnitHealthChanged",					self)
		Apollo.RegisterEventHandler("CombatLogAbsorption", 				"OnUnitHealthChanged",					self)
		
		-- Ability
		--Apollo.RegisterEventHandler("CombatLogInterrupted", 			"OnUnitInterrupted",					self)		
		
		-- Start timer at auto if we have no data
		if not self.strRate then
			self:SetRate("auto")
		end
		
		-- Prep Units
		self:CreateUnitsFromPreload()
		
		-- If nothing broke then we fire the annoucement.
		Event_FireGenericEvent("CombatEvents_Loaded", {version=nVersion})
	end
end

-----------------------------------------------------------------------------------------------
-- CombatEvents Preload
-----------------------------------------------------------------------------------------------
function CombatEvents:OnPreloadUnitCreated(unit)
	-- Cache unit data
	self.unitPreload[unit:GetId()] = unit
end

function CombatEvents:CreateUnitsFromPreload()
	-- Process units created while form was loading
	for _, unit in pairs(self.unitPreload) do
		if unit:IsValid() then
			self:OnUnitCreated(unit)
		end
	end
	self.unitPreload = nil
end


-----------------------------------------------------------------------------------------------
-- CombatEvents Unit Discovery Events
-----------------------------------------------------------------------------------------------
function CombatEvents:OnUnitCreated(unit)
	local unitType = unit:GetType()
	if not unit:ShouldShowNamePlate() and kBlacklistedUnitTypes[unitType] then
		-- Don't track
		return nil
	end
	
	
	local UnitData = {
		unit = unit,
		id = unit:GetId(),
		name = unit:GetName(),
		type = unitType,
		
		isCasting = unit:ShouldShowCastBar(),
	
		-- Vitals
		health = {
			current = unit:GetHealth(),
			max = unit:GetMaxHealth(),
		},
		shield = {
			current = unit:GetShieldCapacity(),
			max = unit:GetShieldCapacityMax(),
		},
		absorb = {
			current = 0,
			max = unit:GetAbsorptionMax() or 0,
		},
	}
	if UnitData.absorb.max then
		UnitData.absorb.current = unit:GetAbsorptionValue()
	end
	
	-- Casting
	if UnitData.isCasting then
		UnitData.cast = {
			type = unit:GetCastBarType(),
			name = unit:GetCastName(),
			pct = unit:GetCastTotalPercent(),
			duration = unit:GetCastDuration(),
			elapsed = unit:GetCastElapsed(),
		}
	end
	
	self.tUnits[UnitData.id] = UnitData
	
	Event_FireGenericEvent("CE_UnitCreated", UnitData)
end


function CombatEvents:OnUnitDestroyed(unit)
	local unitId = unit:GetId()
	if not self.tUnits[unitId] then
		return nil
	end
	self.tUnits[unitId] = nil
	
	Event_FireGenericEvent("CE_UnitDestroyed", {id=unitId})
end

-----------------------------------------------------------------------------------------------
-- CombatEvents OnUnitUpdate
-----------------------------------------------------------------------------------------------
function CombatEvents:OnUnitUpdate()
	for _, UnitData in pairs(self.tUnits) do
		self:UpdateVitals(UnitData)
		self:RefreshCast(UnitData)
	end
end


-----------------------------------------------------------------------------------------------
-- CombatEvents Functions
-----------------------------------------------------------------------------------------------
function CombatEvents:UpdateVitals(UnitData)
	local health = self:RefreshHealth(UnitData)
	local shield = self:RefreshShield(UnitData)
	local absorb = self:RefreshAbsorb(UnitData)
	if health or shield or absorb then
		local event_data = {
			unit = UnitData.unit,
			unitId = UnitData.unitId,
			health = health,
			shield = shield,
			absorb = absorb,
		}
		Event_FireGenericEvent("CE_UnitVitalsChanged", event_data)
	end
end

function CombatEvents:RefreshHealth(UnitData)
	local current = UnitData.unit:GetHealth()

	if UnitData.health.current ~= current then
		UnitData.health.current = current
		UnitData.health.max = UnitData.unit:GetMaxHealth()
		return UnitData.health
	end

end

function CombatEvents:RefreshShield(UnitData)
	local current = UnitData.unit:GetShieldCapacity()

	if UnitData.shield.current ~= current then
		UnitData.shield.current = current
		UnitData.shield.max = UnitData.unit:GetShieldCapacityMax()
		return UnitData.shield
	end
end

function CombatEvents:RefreshAbsorb(UnitData)
	local current = 0
	local max = UnitData.unit:GetAbsorptionMax()
	
	if max and max > 0 and max ~= UnitData.absorb.max then
		local current_new = UnitData.unit:GetAbsorptionValue()
		if UnitData.absorb.current ~= current_new then
			current = current_new
		end
		
		if max ~= UnitData.absorb.max then
			UnitData.absorb.max = max
		end
		UnitData.absorb.current = current
		UnitData.absorb.max = max
		return UnitData.absorb
	end	
end

function CombatEvents:RefreshCast(UnitData)
	local isCasting = UnitData.unit:ShouldShowCastBar()
	if isCasting ~= UnitData.isCasting then
		UnitData.isCasting = isCasting
		local event_data = {
			unit = UnitData.unit,
			unitId = UnitData.unitId,
		}
		if isCasting then
			-- Unit is Casting.
			event_data.is_casting = true
			event_data.type  = UnitData.unit:GetCastBarType()
			event_data.name = UnitData.unit:GetCastName()
			event_data.pct  = UnitData.unit:GetCastTotalPercent()
			event_data.duration = UnitData.unit:GetCastDuration()
			event_data.elapsed = UnitData.unit:GetCastElapsed()
		else
			-- Unit has stopped casting.
			event_data.is_casting = false
		end
		Event_FireGenericEvent("CE_UnitCastStateChanged", event_data)
	end
end

-----------------------------------------------------------------------------------------------
-- CombatEvents CombatLog Events
-----------------------------------------------------------------------------------------------
function CombatEvents:OnUnitHealthChanged(tEventArgs)
	-- Capture unit taking damage and pipe to UpdateVitals
	if tEventArgs.unitTarget then
		local unitId = tEventArgs.unitTarget:GetId()
		local UnitData = self.tUnits[unitId]
		if UnitData then
			self:UpdateVitals(UnitData)
		end
	end
end

--[[
function CombatEvents:OnUnitInterrupted(tEventArgs)
	if tEventArgs then
		for k,v in pairs(tEventArgs) do
			Print(tostring(k).."="..tostring(v))
		end
	end
end
--]]

---------------------------------------------------------------------------------------------------
-- CombatEvents Options
---------------------------------------------------------------------------------------------------
function CombatEvents:OnConfigure()
	for index,info in ipairs(kUpdateTimers) do
		if self.strRate == info.id then
			self.wndRates[index]:SetCheck(true)
			self.wndOptions:FindChild("Body:UpdateRate:DropToggleFPS"):SetText(info.label)
			break
		end
	end
	
	self.wndOptions:Show(true)
	self.wndOptions:ToFront()
end

function CombatEvents:SetRate(strRate)
	local oldRate = self.strRate
	if not oldRate or oldRate ~= strRate then
		self.strRate = strRate
		if oldRate then
			if oldRate == "auto" then
				Apollo.RemoveEventHandler("VarChange_FrameCount",	self)
			else
				Apollo.StopTimer("UpdateTimer_"..oldRate)
			end
		end
		if self.strRate == "auto" then
			Apollo.RegisterEventHandler("VarChange_FrameCount", "OnUnitUpdate",	self)
		else
			Apollo.StartTimer("UpdateTimer_"..self.strRate)
		end
	end
end

---------------------------------------------------------------------------------------------------
-- CombatEvents Saved Data
---------------------------------------------------------------------------------------------------
function CombatEvents:OnSave(eType)
	if eType ~= GameLib.CodeEnumAddonSaveLevel.Account then
		return
	end

	local locWindowLocation = self.wndOptions and self.wndOptions:GetLocation() or self.locSavedOptionsLoc

	local tSaved =
	{
		tWindowLocation = locWindowLocation and locWindowLocation:ToTable() or nil,
		nSavedVersion = knSaveVersion,
		strRate = self.strRate,
	}

	return tSaved
end

function CombatEvents:OnRestore(eType, tSavedData)
	if tSavedData and tSavedData.nSavedVersion == knSaveVersion then
		if tSavedData.tWindowLocation then
			self.locSavedOptionsLoc = WindowLocation.new(tSavedData.tWindowLocation)
		end
		if tSavedData.strRate then
			self:SetRate(tSavedData.strRate)
		end
	end
end


---------------------------------------------------------------------------------------------------
-- OptionsForm Functions
---------------------------------------------------------------------------------------------------

function CombatEvents:OnWindowModeToggle(wndHandler, wndControl)
	if wndHandler ~= wndControl then
		return nil
	end
	wndControl:FindChild("ChoiceContainer"):Show(wndControl:IsChecked())
end

function CombatEvents:OnFPSRadioButton( wndHandler, wndControl, eMouseButton )
	local index = wndControl:GetParent():GetRadioSel("FPSRATE")
	local id = wndHandler:GetData()
	wndControl:GetParent():GetParent():SetText(kUpdateTimers[index].label)
	wndControl:GetParent():GetParent():SetCheck(false)
	wndControl:GetParent():Close()
end


function CombatEvents:OnFPSButtonMouseEnter( wndHandler, wndControl, x, y )
	wndHandler:FindChild("Label_Rate"):Show(true)
end

function CombatEvents:OnFPSButtonMouseExit( wndHandler, wndControl, x, y )
	wndHandler:FindChild("Label_Rate"):Show(false)
end

function CombatEvents:OnAccept( wndHandler, wndControl, eMouseButton )
	local index = self.wndOptions:FindChild("Body:UpdateRate:DropToggleFPS:ChoiceContainer"):GetRadioSel("FPSRATE")
	if index and index > 0 then
		local id = self.wndRates[index]:GetData()
		self:SetRate(id)
	end
	self.wndOptions:Show(false)
end

-----------------------------------------------------------------------------------------------
-- CombatEvents Instance
-----------------------------------------------------------------------------------------------
local CombatEventsInst = CombatEvents:new()
CombatEventsInst:Init()
